__author__ = 'Tyler Hughes'
__credits__ = 'Dr. Adrian Rosebrock'
__version__ = '0.2'

import cv2
import numpy as np
import argparse as ap
import csv

import matplotlib
matplotlib.use('TkAgg')
from matplotlib import pyplot as plt

import time
import collections

argument_parser = ap.ArgumentParser()
argument_parser.add_argument('-i', '--video_file', help='specify an input video file; ignore to use webcam')
argument_parser.add_argument('-c', '--colors', nargs='*', help='specify colors to track; r: red, o: orange, y: yellow, \
    g:green , b: blue, k: black, w: white, e: grey')
argument_parser.add_argument('-o', '--csv_file', help='specify an output csv file; it will be created if it does not exist')
argument_parser.add_argument('-g', '--show_graph', action='store_true', help='show a graph of positions tracked for \
    each color on exit')
argument_parser.add_argument('-rg', '--real_time_graph', action='store_true', help='enable real-time graphing')
argument_parser.add_argument('-r', '--range_of_frames', type=int, nargs=2, help='specify a range of frames to analyze \
    from the video file')
argument_parser.add_argument('-cx', '--crop_x', type=int, nargs=2, help='crop the video stream in the horizontal axis')
argument_parser.add_argument('-cy', '--crop_y', type=int, nargs=2, help='crop the video stream in the vertical axis')
argument_parser.add_argument('-fx', '--flip_x', action='store_true', help='flip the video stream in the horizontal axis')
argument_parser.add_argument('-fy', '--flip_y', action='store_true', help='flip the video stream in the vertical axis')
argument_parser.add_argument('-gb', '--gaussian_blur', action='store_true', help='enable gaussian blur during image processing')

arguments = vars(argument_parser.parse_args())

# These are HSV color bounds.
class ColorBounds:
    RED1_LOWER = np.array([0, 50, 50], np.uint8)
    RED1_UPPER = np.array([9, 255, 255], np.uint8)
    RED2_LOWER = np.array([151, 50, 50], np.uint8)
    RED2_UPPER = np.array([180, 255, 255], np.uint8)
    ORANGE_LOWER = np.array([10, 50, 50], np.uint8)
    ORANGE_UPPER = np.array([15, 255, 255], np.uint8)
    YELLOW_LOWER = np.array([21, 70, 70], np.uint8)
    YELLOW_UPPER = np.array([45, 255, 255], np.uint8)
    GREEN_LOWER = np.array([46, 40, 40], np.uint8)
    GREEN_UPPER = np.array([100, 255, 255], np.uint8)
    BLUE_LOWER = np.array([101, 50, 50], np.uint8)
    BLUE_UPPER = np.array([150, 255, 255], np.uint8)
    BLACK_LOWER = np.array([0, 0, 0], np.uint8)
    BLACK_UPPER = np.array([180, 255, 60], np.uint8)
    WHITE_LOWER = np.array([0, 0, 200], np.uint8)
    WHITE_UPPER = np.array([180, 255, 255], np.uint8)
    GREY_LOWER = np.array([0, 0, 90], np.uint8)
    GREY_UPPER = np.array([180, 255, 180], np.uint8)

class ColorTrack:
    def __init__(self, lower_bound, upper_bound, graph_title, graph_format='k.'):
        self.lower_bounds = [lower_bound]
        self.upper_bounds = [upper_bound]

        self.x_locations = collections.deque()
        self.y_locations = collections.deque()

        self.graph_title = graph_title
        self.graph_format = graph_format

    def create_mask(self, frame):
        mask = np.full((frame.shape[0], frame.shape[1]), 0, dtype=np.uint8)

        for i in range(len(self.lower_bounds)):
            mask = mask | cv2.inRange(frame, self.lower_bounds[i], self.upper_bounds[i])

        mask = cv2.erode(mask, None, iterations=2)
        mask = cv2.dilate(mask, None, iterations=2)

        return mask

    def add_bounds(self, lower_bound, upper_bound):
        self.lower_bounds.append(lower_bound)
        self.upper_bounds.append(upper_bound)
    
    def add_point(self, point):
        self.x_locations.appendleft(point[0])
        self.y_locations.appendleft(point[1])

if arguments.get('video_file', False):
    video_stream = cv2.VideoCapture(arguments['video_file'])
else:
    video_stream = cv2.VideoCapture(0)

if not video_stream.isOpened():
    raise RuntimeError('Could not launch video capture device.')

# Give the capture device time to catch up.
time.sleep(2.0)

# Create the to color tracks.
color_tracks = []

# Create the corresponding color track from the arguments with these functions.
def red_color_track():
    color_track = ColorTrack(ColorBounds.RED1_LOWER, ColorBounds.RED1_UPPER, 'Red Color', 'r.')
    color_track.add_bounds(ColorBounds.RED2_LOWER, ColorBounds.RED2_UPPER)
    color_tracks.append(color_track)

def orange_color_track():
    color_tracks.append(ColorTrack(ColorBounds.ORANGE_LOWER, ColorBounds.ORANGE_UPPER, 'Orange Color', 'y.'))

def yellow_color_track():
    color_tracks.append(ColorTrack(ColorBounds.YELLOW_LOWER, ColorBounds.YELLOW_UPPER, 'Yellow Color', 'y.'))

def green_color_track():
    color_tracks.append(ColorTrack(ColorBounds.GREEN_LOWER, ColorBounds.GREEN_UPPER, 'Green Color', 'g.'))

def blue_color_track():
    color_tracks.append(ColorTrack(ColorBounds.BLUE_LOWER, ColorBounds.BLUE_UPPER, 'Blue Color', 'b.'))

def black_color_track():
    color_tracks.append(ColorTrack(ColorBounds.BLACK_LOWER, ColorBounds.BLACK_UPPER, 'Black Color', 'k.'))

def white_color_track():
    color_tracks.append(ColorTrack(ColorBounds.WHITE_LOWER, ColorBounds.WHITE_UPPER, 'White Color', 'b.'))

def grey_color_track():
    color_tracks.append(ColorTrack(ColorBounds.GREY_LOWER, ColorBounds.GREY_UPPER, 'Grey Color', 'b.'))

color_track_functions = {
    'r': red_color_track,
    'o': orange_color_track,
    'y': yellow_color_track,
    'g': green_color_track,
    'b': blue_color_track,
    'k': black_color_track,
    'w': white_color_track,
    'e': grey_color_track
}

if arguments.get('colors', False):
    for i in range(len(arguments['colors'])):
        color_track_functions[arguments['colors'][i]]()

# Create the frame crop dimensions if specified.
if arguments.get('crop_x', False):
    frame_x_dimensions = [arguments['crop_x'][0], arguments['crop_x'][1]]
else:
    frame_x_dimensions = [0, int(video_stream.get(3))]

if arguments.get('crop_y', False):
    frame_y_dimensions = [arguments['crop_y'][0], arguments['crop_y'][1]]
else:
    frame_y_dimensions = [0, int(video_stream.get(4))]

figure, axes = plt.subplots(1, len(color_tracks))
plt.subplots_adjust(wspace=0.5)

for i in range(len(color_tracks)):
    if len(color_tracks) > 1:
        axes[i].set_title(color_tracks[i].graph_title)
        axes[i].set_xlabel('Position in x (Pixels)')
        axes[i].set_ylabel('Position in y (Pixels)')
    else:
        axes.set_title(color_tracks[i].graph_title)
        axes.set_xlabel('Position in x (Pixels)')
        axes.set_ylabel('Position in y (Pixels)')

def plot_graphs(axes, color_tracks):
    for i in range(len(color_tracks)):
        x_locations = np.array(color_tracks[i].x_locations)
        y_locations = np.array(color_tracks[i].y_locations)

        # If there is only one color track, axes will not have an index.
        if len(color_tracks) > 1:
            axes[i].plot(x_locations.T, y_locations.T, color_tracks[i].graph_format)
        else:
            axes.plot(x_locations.T, y_locations.T, color_tracks[i].graph_format)

show_graph = False
# This flag is used to enable real-time graphing, which is slow.
real_time_graph = False

if arguments.get('show_graph', False):
    show_graph = True

if arguments.get('real_time_graph', False):
    real_time_graph = True

frame_number = 0

while True:
    if arguments.get('range_of_frames', False):
        if frame_number < arguments['range_of_frames'][0]:
            continue

        if frame_number >= arguments['range_of_frames'][1]:
            break

    was_grabbed, frame = video_stream.read()

    if frame is None:
        break

    # Crop the frame.
    frame = frame[frame_y_dimensions[0]:frame_y_dimensions[1], frame_x_dimensions[0]:frame_x_dimensions[1]]

    # Flip the frame if specified.
    if arguments.get('flip_x', False):
        frame = cv2.flip(frame, 1)
    if arguments.get('flip_y', False):
        frame = cv2.flip(frame, 0)

    frame_copy = frame.copy()
    frame_copy = cv2.cvtColor(frame_copy, cv2.COLOR_BGR2HSV)

    if arguments.get('gaussian_blur', False):
        frame_copy = cv2.GaussianBlur(frame_copy, (11, 11), 0)

    mask = np.full((frame_copy.shape[0], frame_copy.shape[1]), 0, dtype=np.uint8)

    for i in range(len(color_tracks)):
        current_mask = color_tracks[i].create_mask(frame_copy)

        _, contours, _ = cv2.findContours(current_mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

        if len(contours) > 0:
            contour = max(contours, key=cv2.contourArea)
            (x, y), radius = cv2.minEnclosingCircle(contour)

            if radius > 0:
                # Draw the minimum enclosing circle.
                cv2.circle(frame, (int(x), int(y)), int(radius), (0, 255, 255), 2)

                # Determine the center of the contour.
                moments = cv2.moments(contour)
                center_point = (int(moments['m10'] / moments['m00']), int(moments['m01'] / moments['m00']))

                # Change the reference point for the y-coordinate of the center to the bottom of the screen.
                flipped_center_point = (center_point[0], abs(center_point[1] - frame_y_dimensions[1]))

                # Draw the center point.
                cv2.circle(frame, center_point, 3, (0, 255, 255), -1)

                color_tracks[i].add_point(flipped_center_point)

        mask = mask | current_mask

    cv2.imshow('mask', mask)
    cv2.imshow('frame', frame)

    if cv2.waitKey(1) & 0xFF == ord('g'):
        real_time_graph = not real_time_graph
    elif cv2.waitKey(1) & 0xFF == ord('q'):
        break

    if real_time_graph:
        plot_graphs(axes, color_tracks)
        plt.pause(0.001)

    frame_number += 1

if show_graph:
    plot_graphs(axes, color_tracks)
    plt.show()

# If specified, write all the position data for each color track to a file.
if arguments.get('csv_file', False):
    with open(arguments['csv_file'], 'w', newline='') as csv_file:
        writer = csv.writer(csv_file)
        
        row = []
        for i in range(len(color_tracks)):
            row.append(color_tracks[i].graph_title)
            row.append('')
        
        writer.writerow(row)
        row.clear()

        for i in range(len(color_tracks)):
            row.append('x')
            row.append('y')
        
        writer.writerow(row)
        row.clear()

        # All color tracks should have the same number of entries - one for each frame analyzed.
        for i in range(len(color_tracks[0].x_locations)):
            for j in range(len(color_tracks)):
                row.append(color_tracks[j].x_locations[i])
                row.append(color_tracks[j].y_locations[i])

            writer.writerow(row)
            row.clear()